//--------------------------------------------------------------------------
//
//   Copyright (c) 2002, Colin Granville
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or
//   without modification, are permitted provided that the following
//   conditions are met:
//
//      * Redistributions of source code must retain the above copyright
//        notice, this list of conditions and the following disclaimer.
//
//      * Redistributions in binary form must reproduce the above
//        copyright notice, this list of conditions and the following
//        disclaimer in the documentation and/or other materials
//        provided with the distribution.
//
//      * The name Colin Granville may not be used to endorse or promote
//        products derived from this software without specific prior
//        written permission.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
//   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
//   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
//   OF THE POSSIBILITY OF SUCH DAMAGE.
//
//--------------------------------------------------------------------------

#include "DocViewToolbar.h"
#include "DocView.h"
#include "Guigfx.h"
#include "UserEvents.h"
#include "Document.h"
#include "outline.h"
#include "GuiHourglass.h"
#include "PSOutputDev.h"
#include "SplashBitmap.h"
#include "Splash.h"
#include "SplashOutputDev.h"
#include "file.h"
#include "GooTimer.h"
#include <ctype.h>

struct TaskData
{
  int          handle;
  const char*  name;
  unsigned int size;
  unsigned int flags;
};

int get_task_handle(const char* task)
{
  int      more=0;
  TaskData data;
  while (!_swix(TaskManager_EnumerateTasks,_INR(0,2) |_OUT(0),more,&data,sizeof(data),&more) && more>0)
  {
    if (!strcmp(task,data.name)) return data.handle;
  }
  return 0;
}

DocViewToolbar::DocViewToolbar(DocView& docView,GuiWindow& ancestor)
 : GuiWindow("toolbar"),
   view(docView),
   page(*this,0xe),
   lastPage(*this,8),
   pageSlider(*this,0xf),
   outline(*this,0x89),
   logicalPageOffset(0),
   currentPage(0),
   pageCount(1),
   oldWidth(-1),
   history(docView,*this,ancestor),
   bookmarks(docView,*this,ancestor),
   showOutlineTarget(&ancestor,User_ShowOutline,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::showOutline),
   gViewPostscriptFileTarget(&ancestor,User_GViewPostscriptFile,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::gViewPostscriptFile),
   gViewPdfFileTarget(&ancestor,User_GViewPdfFile,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::gViewPdfFile),
   adjusterClickedTarget(&ancestor,User_AdjusterClicked,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::adjusterClicked),
   readPageTarget(&ancestor,User_ReadPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::readPage),
   sliderValueChangedTarget(&ancestor,GuiSlider::ValueChanged::Event,
                            this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::sliderValueChanged),
   nextPartPageTarget(&ancestor,User_NextPartPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::nextPage),
   nextPageTarget(&ancestor,User_NextPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::nextPage),
   fastNextPageTarget(&ancestor,User_FastNextPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::nextPage),
   previousPartPageTarget(&ancestor,User_PreviousPartPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::previousPage),
   previousPageTarget(&ancestor,User_PreviousPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::previousPage),
   fastPreviousPageTarget(&ancestor,User_FastPreviousPage,this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::previousPage),
   logicalPageNumberATBSTarget(&ancestor,User_LogicalPageNumberATBS,
                               this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::logicalPageNumberATBS),
   logicalPageNumberSetTarget(&ancestor,User_LogicalPageNumberSet,
                              this,(Claim (DocViewToolbar::*)(GuiToolboxEvent&, const GuiIdBlock&))&DocViewToolbar::logicalPageNumberSet)

{
}

//*************************************************************************

DocViewToolbar::~DocViewToolbar()
{
  Outline* o=view.getDocument().getOutline();
  if (o) o->close(view);
}

//*************************************************************************

void DocViewToolbar::goBack()
{
  const DocPage* page=history.back();
  if (page) setPage(page->page,0);
}

//*************************************************************************

void DocViewToolbar::openWindow(GuiWindow& parent,GuiGetWindowStateBlock& parent_state)
{
  GuiGetWindowStateBlock b;
  getState(b);
  b.behind=parent_state.behind;
  parent_state.behind=b.windowHandle;

  b.visibleArea.xmin=parent_state.visibleArea.xmin;
  b.visibleArea.xmax=b.visibleArea.xmin+parent_state.visibleArea.getWidth();
  int ht=b.visibleArea.getHeight();
  b.visibleArea.ymax=parent_state.visibleArea.ymax;
  b.visibleArea.ymin=b.visibleArea.ymax-ht;
  show(b,&parent);

  int new_width=parent_state.visibleArea.getWidth();
  if (new_width!=oldWidth)
  {
    oldWidth=new_width;
    GuiBBox box;
    pageSlider.getBBox(box);
    box.xmax=new_width-6;
    pageSlider.move(box);
  }
  return;
}
//*************************************************************************

void DocViewToolbar::setFirstPage(int page_ ,int logicalPO, int page_count,bool add_to_history)
{
  if (page_count<=0) page_count=1;
  if (page_ >= page_count) page_=page_count;
  if (page_ <=0) page_=1;

  page.setFocus();
  logicalPageOffset=logicalPO;

  currentPage=page_;
  pageCount=page_count;

  page.setNumber(currentPage+logicalPageOffset);
  lastPage.setNumber(pageCount);

  // crashes if upper==lower
  pageSlider.setBounds( GuiSlider::SetUpper | GuiSlider::SetLower | GuiSlider::SetStepSize,
                        1,(pageCount==1?2:pageCount),1);
  pageSlider.setValue(currentPage);

  Outline* o=view.getDocument().getOutline();
  outline.fade(!(o && o->isOk()));

  if (add_to_history) history.add(currentPage);
}

//*************************************************************************

void DocViewToolbar::setPage(int new_page, bool set_history)
{
  if (new_page>pageCount) new_page=pageCount;
  if (new_page<=0) new_page=1;
  if (new_page==currentPage) return;
  currentPage=new_page;
  if (set_history) history.add(currentPage);
  pageSlider.setValue(currentPage);
  page.setNumber(currentPage+logicalPageOffset);

  view.showPage();
}

//*************************************************************************

Claim DocViewToolbar::showOutline(GuiToolboxEvent&,const GuiIdBlock&)
{
  Outline* o=view.getDocument().getOutline();
  if (o) o->open(view);
  outline.fade(!(o && o->isOk()));
  return CLAIM;
}

//*************************************************************************

Claim DocViewToolbar::gViewPostscriptFile(GuiToolboxEvent&,const GuiIdBlock&)
{
  double timeInMs;
  SplashColor paperColor;

//  if (!getenv("Paint$Dir")) return CLAIM;

  GooTimer msTimer;
  GuiHourglass hourglass;
  char* com="Resources:$.Apps.!Paint.!Run <wimp$ScrapDir>.PDF_SPRITE";
  char* filename=strchr(com,' ')+1;

  {
    msTimer.start();
    paperColor[0] = paperColor[1] = paperColor[2] = 0xff;
    SplashOutputDev out(splashModeRGB8,1,0,paperColor);
    out.startDoc(view.getDocument().getPDFDoc().getXRef());
    view.getDocument().getPage(view.getPage(),&out,90,0);
    out.getBitmap()->writePNMFile(filename);
    msTimer.stop();
    timeInMs = msTimer.getElapsed();
  }
  setFileType(filename,0xff9);
  fprintf(stderr, "Time: %.2f secs\n", timeInMs);
  fflush(stderr);

  int task=get_task_handle("Paint");

  if (task)
  {
    GuiWimpMessage mess;
    mess.data.dataOpen.windowHandle=-2;
    mess.data.dataOpen.padding1=-1;
    mess.data.dataOpen.x=mess.data.dataOpen.y=0;
    mess.data.dataOpen.padding2=0;
    mess.data.dataOpen.fileType=0xff9;
    mess.copyString(mess.data.dataOpen.pathName,filename);
    mess.send(GuiWimp_MDataOpen,task);
  }
  else
  {
    _swix(Wimp_StartTask,_IN(0),com);
  }

  return CLAIM;
}

//*************************************************************************

class GViewMessage : public GuiWimpMessageBase
{
  public:
    enum {MESSAGE=0x4ea04};
    unsigned int actionCode;
    unsigned int pageNumber;
    char         filename[256-sizeof(GuiWimpMessageBase)-12];

};

Claim DocViewToolbar::gViewPdfFile(GuiToolboxEvent& e,const GuiIdBlock& b)
{
  if (!getenv("GView$Dir")) return CLAIM;
//  if (!getenv("GView$Version")) return gViewPostscriptFile(e,b);

  int task=get_task_handle("GView");
  if (task)
  {
    char buffer[1024];
    sprintf(buffer,"<GView$Dir>.!run -page %d %s",view.getPage(),view.getDocument().getFileName().c_str());
    _swix(Wimp_StartTask,_IN(0),buffer);
  }
  else
  {
    GViewMessage mess;
    mess.actionCode=0;
    mess.pageNumber=view.getPage();
    mess.copyString(mess.filename,view.getDocument().getFileName().c_str());
    mess.send(GViewMessage::MESSAGE,task);
  }
  return CLAIM;
}

//*************************************************************************

Claim DocViewToolbar::adjusterClicked(GuiToolboxEvent& ev,const GuiIdBlock& id_block)
{
  if (id_block.self.component==0x87)
  {
    ev.eventCode=User_PreviousPartPage;
    if (gfx::inkey(-1)) ev.eventCode=User_PreviousPage; //shift
    if (gfx::inkey(-2)) ev.eventCode=User_FastPreviousPage; //ctrl
    return previousPage(ev,id_block);
  }

  ev.eventCode=User_NextPartPage;
  if (gfx::inkey(-1)) ev.eventCode=User_NextPage; //shift
  if (gfx::inkey(-2)) ev.eventCode=User_FastNextPage; //ctrl
  return nextPage(ev,id_block);
};

//*************************************************************************

Claim DocViewToolbar::readPage(GuiToolboxEvent&,const GuiIdBlock&)
{
  setPage(page.getNumber()-logicalPageOffset);
  view.scroll(DocView::SCROLL_TOP);
  return CLAIM;
};

//*************************************************************************

Claim DocViewToolbar::sliderValueChanged(GuiToolboxEvent& ev,const GuiIdBlock&)
{
  GuiSlider::ValueChanged& s= (GuiSlider::ValueChanged&)ev;

  if (s.flags == GuiSlider::ValueChanged::DragStart) history.addPending();
  if (s.newValue<=pageCount && currentPage!=s.newValue)
  {
    currentPage=s.newValue;
    page.setNumber(currentPage+logicalPageOffset);
    view.showPage();
    history.addPending(currentPage);
  }
  return CLAIM;
}

//*************************************************************************

Claim DocViewToolbar::nextPage(GuiToolboxEvent& ev,const GuiIdBlock&)
{

  if (ev.eventCode==User_NextPartPage && view.scroll(DocView::SCROLL_DOWN)) return CLAIM;

  int page=currentPage;
  setPage(currentPage+(ev.eventCode==User_FastNextPage?10:1));
  if (page!=currentPage) view.scroll(DocView::SCROLL_TOP);

  return CLAIM;
}

//*************************************************************************

Claim DocViewToolbar::previousPage(GuiToolboxEvent& ev,const GuiIdBlock&)
{
  if (ev.eventCode==User_PreviousPartPage && view.scroll(DocView::SCROLL_UP)) return CLAIM;

  int page=currentPage;
  setPage(currentPage-(ev.eventCode==User_FastPreviousPage?10:1));
  if (page!=currentPage)
     view.scroll(ev.eventCode==User_PreviousPartPage ? DocView::SCROLL_BOTTOM : DocView::SCROLL_TOP);
  return CLAIM;
}

//*************************************************************************

Claim DocViewToolbar::logicalPageNumberATBS(GuiToolboxEvent&,const GuiIdBlock& idBlock)
{
  GuiWindow win(idBlock.self.id);
  NumberField(win,0).setNumber(currentPage+logicalPageOffset);
  return CLAIM;
}

//*************************************************************************

Claim DocViewToolbar::logicalPageNumberSet(GuiToolboxEvent&,const GuiIdBlock& idBlock)
{

  GuiWindow win(idBlock.self.id);
  logicalPageOffset = NumberField(win,0).getNumber()-currentPage;
  page.setNumber(currentPage+logicalPageOffset);

  return CLAIM;
}
